﻿
Imports System.Collections.ObjectModel
Imports System.IO
Imports WpfNotes.BaseClasses
Imports WpfNotes.Helper.IO
Imports WpfNotes.Model

Namespace Context
    Public Class NotesContext
        Inherits NotifyBase
        Implements IDisposable


        Private ReadOnly _xmlPath As String

        'Pessimistic Lock und letzter Abrufzeitpnkt
        Private Const LockFileName As String = "pessimisticlock.loc"
        Private _lastSyncTime As Date

        Private _xmlFileWatcher As FileSystemWatcher
        Private Property CurrentContext As XmlContext

        'Events
        Friend Event DataFileWasChanged()
        Friend Event ErrorMessageSent(message As String)


        Public Sub New()
            If String.IsNullOrEmpty(My.Settings.NotesXmlPath) Then Throw New Exception("Der Pfad zum XML-Datenfile muss in den Settings hinterlegt sein.")
            If Not File.Exists(My.Settings.NotesXmlPath) Then Throw New Exception("Der Pfad zur XML-Datendatei aus den Settings existiert nicht.")
            _xmlPath = My.Settings.NotesXmlPath
            InitNotesContext()
        End Sub

        Public Sub New(xmlPath As String, Optional savePathToSettings As Boolean = False)
            If Not Directory.Exists(Path.GetDirectoryName(xmlPath)) Then Throw New Exception("Der Pfad zum XML-Datendateiordner existiert nicht.")
            _xmlPath = xmlPath
            If savePathToSettings Then My.Settings.NotesXmlPath = _xmlPath : My.Settings.Save()
            InitNotesContext()
        End Sub

        Private Sub InitNotesContext()
            LoadContext()
            SetupFileWatcher()
        End Sub

        Friend Sub LoadContext()
            Dim ser As New XmlSerializer()
            CurrentContext = ser.DeSerialize(Of XmlContext)(_xmlPath, New XmlContext)
            _lastSyncTime = DateTime.Now
        End Sub

#Region "FileSystemWatcher"

        Private Sub SetupFileWatcher()
            _xmlFileWatcher = New FileSystemWatcher(Path.GetDirectoryName(_xmlPath))
            AddHandler _xmlFileWatcher.Changed, AddressOf DataFileChanged
            _xmlFileWatcher.Filter = Path.GetFileName(_xmlPath)
            _xmlFileWatcher.EnableRaisingEvents = True
        End Sub

        Private Sub DataFileChanged(sender As Object, e As FileSystemEventArgs)
            If e.ChangeType = WatcherChangeTypes.Changed Then
                RaiseEvent DataFileWasChanged()
            End If
        End Sub

#End Region


#Region "Die Properties welche quasi die Tabellen representieren"


        Public ReadOnly Property Notes As ObservableCollection(Of Note)
            Get
                Return CurrentContext.Notes
            End Get
        End Property

        Public ReadOnly Property Categories As ObservableCollection(Of Category)
            Get
                Return CurrentContext.Categories
            End Get
        End Property

#End Region


        Public ReadOnly Property IsDataUpToDate As Boolean
            Get
                Dim fi = New FileInfo(_xmlPath)
                Return fi.LastWriteTime <= _lastSyncTime
            End Get
        End Property

        Public ReadOnly Property HasPessimisticLock As Boolean
            Get
                Return File.Exists($"{Path.GetDirectoryName(_xmlPath)}\{LockFileName}")
            End Get
        End Property



        Public Function SaveChanges() As Boolean
            If HasPessimisticLock Then RaiseEvent ErrorMessageSent("Die pessimistische Sperre wurde verhängt, so ist ein speichern nicht möglich.")
            Try
                _xmlFileWatcher.EnableRaisingEvents = False
                LockForChanges()

                Dim ser As New XmlSerializer()
                ser.Serialize(Of XmlContext)(_xmlPath, CurrentContext)
                _lastSyncTime = DateTime.Now
                Return True
            Catch ex As Exception
                RaiseEvent ErrorMessageSent(ex.Message)
                Return False
            Finally
                UnlockForChanges()
                _xmlFileWatcher.EnableRaisingEvents = True
            End Try

        End Function




        Public Sub LockForChanges()
            Using fs As New FileStream($"{Path.GetDirectoryName(_xmlPath)}\{LockFileName}", FileMode.Create)
                fs.Close()
            End Using
        End Sub

        Public Sub UnlockForChanges()
            File.Delete($"{Path.GetDirectoryName(_xmlPath)}\{LockFileName}")
        End Sub


        Friend Sub Seed()
            LockForChanges()

            If Not Categories.Any Then
                Categories.Add(New Category() With {.Title = "Normale TODOs", .Desc = "Die Kategorie welche ganz normale Todos enthält"})
                Categories.Add(New Category() With {.Title = "Wichtige TODOs", .Desc = "Die Kategorie welche wichtige Erledigungen enthält"})
                Categories.Add(New Category() With {.Title = "Business TODOs", .Desc = "Geschäftliche TODO'S"})
            End If
            If Not Notes.Any Then
                Notes.Add(New Note() With {.Title = "Test Nr.1", .Content = "Content Nr. 1 mit Text", .CategorieId = Categories.First.ID})
                Notes.Add(New Note() With {.Title = "Test Nr.2", .Content = "Content Nr. 2 mit Text", .CategorieId = Categories.First.ID, .Finished = True})
                Notes.Add(New Note() With {.Title = "Test Nr.3", .Content = "Content Nr. 3 mit Text", .CategorieId = Categories.First.ID})
            End If
            UnlockForChanges()
            SaveChanges()
        End Sub


#Region "IDisposable Support"
        Private _disposedValue As Boolean

        ' IDisposable
        Protected Overridable Sub Dispose(disposing As Boolean)
            If Not _disposedValue Then
                If disposing Then
                    UnlockForChanges()
                    CurrentContext = Nothing
                    RemoveHandler _xmlFileWatcher.Changed, AddressOf DataFileChanged
                End If

            End If
            _disposedValue = True
        End Sub

        Public Sub Dispose() Implements IDisposable.Dispose
            Dispose(True)
        End Sub
#End Region

    End Class


    <Serializable>
    Public Class XmlContext
        Inherits NotifyBase

        Public Sub New()
            Notes = New ObservableCollection(Of Note)
            Categories = New ObservableCollection(Of Category)
        End Sub

        Private _allNotes As ObservableCollection(Of Note)
        Public Property Notes() As ObservableCollection(Of Note)
            Get
                Return _allNotes
            End Get
            Set(ByVal value As ObservableCollection(Of Note))
                _allNotes = value
                RaisePropertyChanged()
            End Set
        End Property


        Private _allCategories As ObservableCollection(Of Category)
        Public Property Categories() As ObservableCollection(Of Category)
            Get
                Return _allCategories
            End Get
            Set(ByVal value As ObservableCollection(Of Category))
                _allCategories = value
                RaisePropertyChanged()
            End Set
        End Property

    End Class




End Namespace